/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	To check and demonstrate utlgac module	*/
/*	usage.					*/
/*						*/
/************************************************/
#include	<dlfcn.h>
#include	<unistd.h>
#include	<errno.h>
#include	<syslog.h>
#include	<string.h>
#include	<stdio.h>
#include	<signal.h>
#include	<time.h>
#include	<sys/resource.h>
#include	<sys/prctl.h>
#include 	<malloc.h>

#include	<utlgac.h>

#define	KRADIR	"/var/crash"

enum bool {false,true};

static	char 	*appname="yagac";
static	int	debug=0;
static	int	verbose=false;
static	char	*base="";
/*

*/
/************************************************/
/*						*/
/*	Procedure to display the program usage	*/
/*						*/
/************************************************/
static void usage()

{
(void) fprintf(stderr,"%s Version %s\n",appname,gac_getversion());
(void) fprintf(stderr,"usage:  %s "
		      "[-b base] "
		      "[-d debug] "
		      "[-h ] "
		      "[-v ] "
		      "\n",appname);
(void) fprintf(stderr,"\t\t-b base      : application working directory\n");
(void) fprintf(stderr,"\t\t-c nbrchkr   : override default checker number\n");
(void) fprintf(stderr,"\t\t-d debug     : debug level\n");
(void) fprintf(stderr,"\t\t-h           : display usage\n");
(void) fprintf(stderr,"\t\t-v           : verbose debug\n");
(void) exit(-1);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to allow process core DUMP	*/
/*	to trace origin of problem		*/
/*						*/
/************************************************/
void allow_core_dump()

{
struct rlimit limites;

if (getrlimit(RLIMIT_CORE,&limites)<0) {
  (void) fprintf(stderr,"getrlimit error='%s'",strerror(errno));
  }
limites.rlim_cur=limites.rlim_max;
if (setrlimit(RLIMIT_CORE,&limites)<0) {
  (void) fprintf(stderr,"setrlimit error='%s'",strerror(errno));
  }
(void) prctl(PR_SET_DUMPABLE,1,0,0,0);/*to allow core-dump	*/
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to display an core_dump	*/
/*	message and terminate application	*/
/*						*/
/************************************************/
void do_core_dump(const char *crashdir,const char *fmt,va_list args)

{
#define MXSTEP 30

va_list ap;
char *str;

va_copy(ap,args);
if (chdir(crashdir)<0) {
  char command[2000];
  char dirdump[1000];

  (void) snprintf(dirdump,sizeof(dirdump),"%s/coredump/","/tmp");
  (void) snprintf(command,sizeof(command),"mkdir -p %s",dirdump);
  (void) system(command);
  if (chdir(dirdump)<0) {
    (void) abort();
    }
  crashdir=dirdump;
  }
(void) vasprintf(&str,fmt,ap);
(void) syslog(LOG_INFO,str);
(void) syslog(LOG_INFO,"going to CORE DUMP (in %score.%d)\n",
		       crashdir,getpid());
(void) sleep(5);	/*to avoid crash avalanche	*/
(void) abort();		/*doing to do the abort		*/
(void) exit(-1);	/*Theoriticaly unreachable	*/
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to display an alert in the	*/
/*	debug log.				*/
/*						*/
/************************************************/
void go_alert(const int dlevel,const char *fmt,...)

{
if (debug>=dlevel)
  {
  va_list args;
  char *strloc;

  (void) gac_inhook(false);
  va_start(args,fmt);
  (void) vasprintf(&strloc,fmt,args);
  if (verbose==true) {
    time_t curtime;
    char logtime[30];

    curtime=time((long *)0);
    (void) fflush(stderr);
    (void) strftime(logtime,sizeof(logtime),"%H:%M:%S",localtime(&curtime));
    (void) fprintf(stderr,"%s %s\n",logtime,strloc);
    (void) fflush(stderr);
    }
  else {
    (void) syslog(LOG_INFO,"%s",strloc);
    }
  (void) free(strloc);
  va_end(args);
  (void) gac_inhook(true);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to prepare a core_dump in	*/
/*	the right directory.			*/
/*						*/
/************************************************/
void go_core_dump(char *frmt,...)

{
va_list args;
char *crashinfo;

va_start(args,frmt);
(void) vasprintf(&crashinfo,frmt,args);
(void) go_alert(1,crashinfo);
(void) free(crashinfo);
(void) do_core_dump(KRADIR,frmt,args);
va_end(args);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to catch signal and do what is*/
/*	needed.					*/
/*						*/
/************************************************/
static void trapped(int sig)

{
#define	OPPB	"yagac.c:trapped,"

switch (sig) {
  case SIGSEGV   :
    (void) go_core_dump("Application genuine memory violation");
    break;
  default        :
    (void) go_core_dump("%s Unexpected signal <%s> received",
			OPPB,sys_siglist[sig]);
    break;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to set the SIGV signal trap	*/
/*	purpose is to CORE_DUMP in case we have	*/
/*	a memory failure of some kind.		*/
/*						*/
/************************************************/
static void trapsegv(int onoff)

{
static struct sigaction oldsa;

if (onoff==true) {
  struct sigaction newsa;

  newsa.sa_flags=0;
  newsa.sa_handler=trapped;
  (void) sigemptyset(&newsa.sa_mask);
  (void) sigaction(SIGSEGV,&newsa,&oldsa);
  }
else {
  (void) sigaction(SIGSEGV,&oldsa,(struct sigaction *)0);
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to assign memory, according	*/
/*	the step  memory s assigned in different*/
/*	part of the code.			*/
/*	Such the test show a good example of	*/
/*	leaking report.				*/
/*						*/
/************************************************/
static void *leakage(int step,int size,u_long *leaked)

{
#define MSG	25
void *ptr;
int taille;
char *seq;

ptr=(void *)0;
taille=MSG+size+1;
(void) asprintf(&seq,"%02d Leaking %03d bytes",step,taille);
switch (step) {
  case 0	:
    ptr=calloc(taille,sizeof(char));
    break;
  default	:
    ptr=malloc(taille);
    break;
  }
(void) strncpy(ptr,seq,taille-1);
(void) free(seq);
(*leaked)+=taille;
return ptr;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to assign memory, but the	*/
/*	code is not very good and a leak is done*/
/*	by purpose.				*/
/*						*/
/************************************************/
static char **leaking(u_long *leaked)

{
#define	MXMEM	5

char **ptr;
int i;

*leaked=(u_long)0;
ptr=(char **)calloc(1,sizeof(char *));
for (i=0;i<MXMEM;i++) {
  void *leak;

  ptr=realloc(ptr,(i+2)*sizeof(char *));
  (void) asprintf(ptr+i,"Assigning good memory num=%03d",i);
  leak=leakage(0,4*i,leaked);
  leak=leakage(1,8*i,leaked);
  ptr[i+1]=(char *)0;
  }
return ptr;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to open a garbage collection	*/
/*	section.				*/
/*						*/
/************************************************/
static GACTYP *gacopen(int tgtdbg,char *callername)

{
GACTYP *gat;

if ((gat=gac_open(tgtdbg,appname,callername))==(GACTYP *)0) {
  switch (errno) {
    case 0	:	/*no GAC opened	*/
      break;
    default	:	/*report error	*/
      (void) go_alert(0,"procedure %s is not able to open GAC (errno=<%s>)",
			callername,strerror(errno));
      break;
    }
  }
return gat;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to close a garbage collection	*/
/*	section.				*/
/*						*/
/************************************************/
static long gacclose(GACTYP *gat,char *callername)

{
long remaining;

switch ((gac_close(gat,&remaining))) {
  case -1	:	/*report trouble	*/
    (void) go_alert(0,"Procedure %s error closing GAC (errno=<%s>)",callername,strerror(errno));
    remaining=0;
    break;
  case  0	:	/*no trouble		*/
    (void) go_alert(0,"According malloc, no memory found still assigned, good");
    break;
  case  1	:	/*malloc trouble	*/
    (void) go_alert(0,"According malloc procedure %s has still '%ld' bytes assigned in memory",callername,remaining);
    break;
  }
return remaining;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Forcing memory leak and trashing leaked	*/
/*	memory while keeping 'good' memory.	*/
/*						*/
/************************************************/
static void test_trashingleak(int tstdbg)

{
GACTYP *gat;
long remaining;
int phase;
int proceed;

gat=(GACTYP *)0;
remaining=0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*need to do test?	*/
      if (debug<tstdbg) {
        (void) go_alert(0,"Testing 'trashing leak' is done only if debug level is '%d' or above,",tstdbg);
        (void) go_alert(0,"Current debug level is '%d'.",debug);
	phase=999;	/*no need to go further	*/
        }
      break;
    case 1	:	/*always openning a GC	*/
      (void) go_alert(0,"Testing 'trashing leak' extracting good memory from leaked memory");
      gat=gacopen(0,"extractor");
      break;
    case 2	:	/*getting + leak memory	*/
      if (gat!=(GACTYP *)0) {	/*always	*/
	GACTYP *locgat;
	char **ptr;
	u_long leaked;
      
        locgat=gacopen(0,"leaker");
	if ((ptr=leaking(&leaked))!=(char **)0) {
          char **gptr;
	  int i;

          /*moving down to previous GC		*/
          (void) gac_gear(gear_down,(long *)0);
          (void) go_alert(0,"Sorting leaked memory from good memory");
	  gptr=(char **)calloc(1,sizeof(char *));
	  for (i=0;ptr[i]!=(void *)0;i++) {
	    gptr=(char **)realloc(gptr,(i+2)*sizeof(char *));
	    gptr[i]=strdup(ptr[i]);
	    gptr[i+1]=(char *)0;
	    }
          (void) gac_gear(gear_up,(long *)0);
	  for (i=0;ptr[i]!=(void *)0;i++) 
	    (void) free(ptr[i]); 
	  (void) free(ptr);
	  ptr=gptr;
	  }
	switch (gac_close(locgat,&remaining)) {
	  case -1	:
	    (void) go_alert(0,"gac_close detected '%d' leaked memory bytes (expected '%d' bytes leaked), discarding them",remaining,leaked);
	    (void) gac_trash(tstdbg+2,locgat,&remaining);
	    break;
	  case  0	:
	    (void) go_alert(0,"Unable to find the leaked memory! (trouble with GAC?)");
	    break;
	  default 	:
	    (void) go_alert(0,"Gac didn't detect leaked memory, but malloc is reporting '%ld' bytes not accounted for",remaining);
	    break;
	  }
	/*cleaning previous memory now		*/
	if (ptr!=(char **)0) {
	  int i;

	  for (i=0;ptr[i]!=(void *)0;i++) 
	    (void) free(ptr[i]); 
	  (void) free(ptr);
	  ptr=(char **)0;
	  }
        }
      break;
    case 3	:	/*reporting leak	*/
      switch (gac_trash(tstdbg,gat,&remaining)) {
	case -1	:
	  (void) go_alert(0,"According malloc there is still '%d' bytes allocated",remaining);
	  (void) go_alert(0,"This mean gac_trash is not working properly or glibc assigned memory for its own purpose");
	  break;
	case  0	:
	  (void) go_alert(0,"No memory leaked anymore (good)");
	  break;
	default	:
	  (void) go_alert(0,"Gac didn't detect leaked memory, but malloc is reporting '%ld' bytes still not accounted for",remaining);
	  break;
	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Forcing memory leak and returning report*/
/*						*/
/************************************************/
static void test_leakdetector(int tstdbg)

{
GACTYP *gat;
long remaining;
int phase;
int proceed;

gat=(GACTYP *)0;
remaining=0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*need to do test?	*/
      if (debug<tstdbg) {
        (void) go_alert(0,"Testing 'leak detector' is done only if debug level is '%d' or above,",tstdbg);
        (void) go_alert(0,"Current debug level is '%d'.",debug);
	phase=999;	/*no need to go further	*/
        }
      break;
    case 1	:	/*always openning a GC	*/
      (void) go_alert(0,"Testing 'leak detector' create leaked memory and report about it");
      gat=gacopen(0,"leakdetector");
      break;
    case 2	:	/*getting + leak memory	*/
      if (gat!=(GACTYP *)0) {	/*always	*/
	char **ptr;
	u_long leaked;
      
	if ((ptr=leaking(&leaked))!=(char **)0) {
	  int i;

	  for (i=0;ptr[i]!=(char *)0;i++) 
	    (void) free(ptr[i]); 
	  (void) free(ptr);
	  }
        }
      break;
    case 3	:	/*reporting leak	*/
      if (gac_trash(tstdbg,gat,&remaining)<0) {
	(void) go_alert(0,"According malloc there is still '%d' bytes allocated",remaining);
	(void) go_alert(0,"This mean gac_trash is not working properly or glibc assigned memory for its own purpose");
	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Forcing faulty condition		*/
/*						*/
/*	Procedure to check allocation error	*/
/*	detection, realloc unknown memory, 	*/
/*	free memory not known by garbage	*/
/*	collector, allocated pointer already	*/
/*	known.					*/
/*						*/
/************************************************/
static void test_allocerror(int tstdbg)

{
void *ptr;
long remaining;
int phase;
int proceed;

ptr=(void *)0;
remaining=(long)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*do we need to do test?*/
      if (debug<tstdbg) {
        (void) go_alert(0,"Testing 'Allocation error' is done only if debug level is '%d' or above,",tstdbg);
        (void) go_alert(0,"Current debug level is '%d'.",debug);
	phase=999;	/*no need to go further	*/
	}
      break;
    case 1	:	/*check already known	*/
      ptr=strdup("Memory pointer to be seen twice by garbage Collector");
      if (ptr!=(char *)0) {	/*always	*/
        char *ptrx;

        (void) gac_inhook(false);
        (void) free(ptr);
        (void) gac_inhook(true);
        ptrx=strdup("Memory pointer to be seen twice by garbage Collector");
	if (ptr==ptrx) 
          (void) gac_report(tstdbg,"Previous message \"...memory 'already set' ptr='%p'\" was expected\n",ptr);
	else {
          (void) gac_report(tstdbg,"Unable to test \"...memory 'already set' ptr='%p'\" condition\n",ptr);
	  (void) free(ptr);
	  }
	ptr=ptrx;
	}
      (void) free(ptr);
      break;
    case 2	:	/*realloc unknown memory*/
      (void) gac_inhook(false);
      ptr=strdup("Memory pointer not to be seen by Garbage collector");
      (void) gac_report(tstdbg,"Next message \"...Unable to find 'realloc memory' previous ptr='%p'\" is expected\n",ptr);
      (void) gac_inhook(true);
      ptr=realloc(ptr,strlen(ptr)+20);
      (void) free(ptr);
      break;
    case 3	:	/*free unknown memory	*/
      (void) gac_inhook(false);
      ptr=strdup("Memory pointer not to be seen freed twice by garbage Collector");
      (void) gac_report(tstdbg,"Next message \"...Unable to find 'to be freed memory' ptr='%p'\" is expected\n",ptr);
      (void) gac_inhook(true);
      (void) free(ptr);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to perform all test		*/
/*	an give example how to use yagac.	*/
/*						*/
/************************************************/
static void testing()

{
#define DBGTST	3

GACTYP *gat;
long remaining;
int proceed;
int phase;

gat=(GACTYP *)0;
remaining=(long)0;
phase=0;
proceed=true;
(void) go_alert(DBGTST,"%s","Starting testing");
while (proceed==true) {
  switch (phase) {
    case 0	:	/*openning a GC only if	*/
      gat=gacopen(DBGTST,"testing");
      break;
    case 1	:	/*test allocation error	*/
      (void) test_allocerror(DBGTST+1);
      break;
    case 2	:	/*test leak detection	*/
      (void) test_leakdetector(DBGTST);
      break;
    case 3	:	/*test trashing memory	*/
      (void) test_trashingleak(DBGTST-1);
      break;
    default	:	/*SAFE Guard		*/
      (void) gacclose(gat,"testing");
      proceed=false;
      break;
    }
  phase++;
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Program Entry.				*/
/*						*/
/************************************************/
int main(int argc,char *argv[])

{
char c;

(void) trapsegv(true);
(void) allow_core_dump();
(void) openlog(appname,LOG_PID,0);
base=strdup(base);
while ((c=getopt(argc,argv,"b:d:hv"))!=EOF) {
  switch(c) {
    case   'b'  :
      (void) free(base);
      base=strdup(optarg);
      break;
    case   'd'  :
      debug=atoi(optarg);
      break;
    case   'v'  :
      verbose=true;
      break;
    case   'h'  :
    default	:
      (void) usage();
      (void) exit(-1);
      break;
    }
  }
(void) gac_setcurbase(&debug,&verbose,(char *)0);
(void) testing();
(void) closelog();
(void) trapsegv(false);
return 0;
}
